// SPDX-FileCopyrightText: 2022 Linus Jahn <lnj@kaidan.im>
//
// SPDX-License-Identifier: LGPL-2.1-or-later

///
/// \class QXmppMessageHandler
///
/// Interface for handling messages.
///
/// \since QXmpp 1.5
///

///
/// \fn QXmppMessageHandler::handleMessage()
///
/// Handles the message.
///
/// \returns Whether the message has been handled and no other extensions should process it.
///
